function Output = GLRRadj(X,W,Y0,n,r,MCMCpara,TypeCov)
%Generalized Reduced Rank Regression Via Decomposition
%
% Model: Y0 = U*(Delta.*Z)*V + WB + E
%-----------------------------------------------------------------
%
%USAGE: Output = Bayesian Longitudinal Low Rank Regression results
%INPUT :  
%   Y0: N x d, input data matrix.  every column is a rsponse from the same subject.
%   X: N x p, covariate matrix for low rank component.
%   W: N x q, prognostic factor and spline matrix for traditiona component.

%   Theta0: a struct data including the initial parameters of the model. The parameters of the model are initialized as follows,
%         U = Theta0.U: p x r matrix as a dictionary for lowrank learning
%         V = Theta0.V: r x d coefficient matrix
%         Delta = Theta0.Delta: r x r diagnal matrix for weighting Z.
%         G = Coefficients for Splines and other Covariates
%  
%         gamma = Theta0.gamma: scalar, precision of the noise.
%
%   hyperpara: MCMC hyperparameters
%   MCMCpara: number of MCMC samples
%       nBurnin: scalar, number of burnin iterations
%       nCollect: scalar, number of collected samples%
%
%
%OUTPUT:
%   Output: a struct data.
        % Output.PostTau_e = Posterior mean of tau_e, error precision;

        % Output.PostB = Posterior mean of B, regression coeficients;
        % Output.Bsd = Posterior standard deviation of B, regression coeficients;

        % Output.PostDelta = Posterior mean of Delta;
        % Output.PostTau_del = Posterior mean of tau_del, precison of delta;
        % Output.PostU = Postosterior mean of U;
        % Output.Usd = Posterior standard deviation of U;
        % Output.PostV = Posterior mean of V;
        % Output.Vsd = Posterior standard deviation of V;

        % Output.nBurnin = nBurnin;
        % Output.nCollect = N;

% ---------------------
% check input arguments
% ---------------------

if nargin < 6
    r=3;
end

if nargin < 7
    nBurnin=1e4;
    nCollect=1e4;
end

if nargin < 8
    TypeCov = 'iid';
end
%
if isempty(MCMCpara)
   nBurnin = 1e4;
   nCollect = 1e4;
end

nBurnin = MCMCpara.nBurnin;
nCollect = MCMCpara.nCollect;
%%
% intitialize hyperparameters

a0 = MCMCpara.a0; c0 = MCMCpara.c0; e0 = MCMCpara.e0; g0 = MCMCpara.g0;
b0 = MCMCpara.b0; d0 = MCMCpara.d0; f0 = MCMCpara.f0; h0 = MCMCpara.h0;

% get dimensions        
[N,p] = size(X);
d = size(Y0,2);
q = size(W,2);

% intitialize mean parameters
U = ones(p,r)*0.1; % low-rank component
V = ones(d,r)*0.1; % low-rank component
Delta = ones(r,1); % low-rank component
G = zeros(q,d); % polynomial and spline coefficients
%s = 3; % for cubic spline

% intitialize precision parameters
O = eye(d); % unstructured covariance component
D_G = zeros(q,q); % perior precision of G
tau_del = 1;
tau_e = 1; % needed for iid case instead of O
tau_G = 1;
tau_b = 1;
tau_u = 1;
Tau_v = ones(d,1);
    
% pre-computing some matrix products
XX = X'*X;
WW = W'*W;

Id = eye(d);
Ip = eye(p);
Iq = eye(q);
In = eye(n);
% initialize outputs;
SumU = zeros(p,r);
SumSU = zeros(p,r);
SumV = zeros(d,r);
SumSV = zeros(d,r);
SumDelta = zeros(r,1);
SumSDelta = zeros(r,1);
SumB = zeros(p,d);
SumSB = zeros(p,d);
SumG = zeros(q,d);
SumSG = zeros(q,d);
%Sumb = zeros(n,d);
%SumSb = zeros(n,d);
SumO = zeros(d,d);
SumSO = zeros(d,d);

%%
for iter = 1:(nBurnin + nCollect)
%%  Sampling the B = U \Delta V 
    B = U*diag(Delta)*V';
    Y = Y0 - X*B - W*G ;  % subtracting all the layers
%     %---------------------------------------------------------------
%     % sampling by LAYER
%     %---------------------------------------------------------------
    for l = 1:r
        Y = Y + X*Delta(l)*U(:,l)*V(:,l)'; % adding back the l-th layer
        XY = X'*Y;
        
        %Sample U / U----------------------------------------------------------
        %sig_Ul = inv((p*Ip + (Delta(l)).^2*(V(:,l)'*O*V(:,l)*XX)));
        %if n>p
            %sig_Ul = inv((tau_u*eye(p) + (Delta(l)).^2*(V(:,l)'*O*V(:,l)*XX)));
            sig_Ul = (tau_u*eye(p) + (Delta(l)).^2*(V(:,l)'*O*V(:,l)*XX))\eye(p);
        %else
        %    tmp1 = ((Delta(l).^2*(V(:,l)'*O*V(:,l)))^(-1)*eye(n)+X*X'/tau_u);
        %    tmp2 = chol(tmp1,'lower')\X;
        %    tmp3 = tmp2'*tmp2;
        %    %tmp3 = X'*linsolve(tmp1,X);            
        %    sig_Ul = (eye(p)-tmp3/tau_u)/tau_u;
        %end
        mu_Ul = (Delta(l))*sig_Ul*(XY*O*V(:,l));
        U(:,l) = mvnrnd(mu_Ul, sig_Ul);
                
        %Vample V / V ----------------------------------------------------------
    
        %sigVl = inv(d*Id + (Delta(l)).^2*U(:,l)'*XX*U(:,l)*O);        
        sigVl = inv(diag(Tau_v) + (Delta(l)).^2*U(:,l)'*XX*U(:,l)*O);
        muVl = Delta(l)*(sigVl*(O*XY'*U(:,l)))';
        V(:,l) = mvnrnd(muVl, sigVl);
        
        %Sample Delta(l)
        sig_Delta = 1/(tau_del + (U(:,l)'*XX*U(:,l))*(V(:,l)'*O*V(:,l)));
        mu_Delta = sig_Delta*U(:,l)'*XY*O*V(:,l);
        Delta(l) = normrnd(mu_Delta, sqrt(sig_Delta));
        %------------------------------------------------------------------
        
        Y = Y - X*Delta(l)*U(:,l)*V(:,l)'; % subtracting the l-th layer again
    end  
    B = U*diag(Delta)*V';
% Sampling spine coefficients G one columns at a time
    Y = Y + W*G;
    for k = 1:d
        Sig_Gk = inv(WW*O(k,k)+tau_G*Iq);
        G_k = DropColumn(G,k);
        Y_k = DropColumn(Y,k);
        Ok = SliceColumn(O,k);
        G(:,k) = (mvnrnd(Sig_Gk*W'*(Y(:,k)*O(k,k)+(Y_k - W*G_k)*Ok), Sig_Gk));
    end;
    E = Y - W*G;
    
%    %---------------------------------------------------------------
    % sampling the precision parameters
    %%%---------------------------------------------------------------
%     E = Y0 - X*B - W*G;
    tau_del = gamrnd(a0+r/2,1./(b0 + sum(Delta.^2)/2));
    %tau_G = gamrnd(g0+0.5*(q-s-1)*d,1./(h0 + sum(sum(G.^2)/2)));
    tau_G = gamrnd(g0+0.5*(q)*d,1./(h0 + sum(sum(G.^2)/2)));
    
    
    %Sample tau_u 
    tau_u = gamrnd(p*r/2+c0, 1./(d0 + 0.5*sum(sum(U.*U))));
    %Sample Tau_v 
    Tau_v = gamrnd(0.5*r*ones(d,1), 1./(f0 + 0.5*(sum(V.*V,2))));
    
    
    if strcmp('Wish', TypeCov) %(Type=='Wish') % Covariance estimation Type
        Scale = inv(E'*E + 1\O_0);
        df = m_0 + n - d - 1;
        O = wishrnd(Scale,df);
    elseif strcmp('Diag', TypeCov)
       nS = E'*E;
       O = eye(d);
       for k = 1:d;
          O(k,k) = max(1e-3, gamrnd((alpha0+n/2), 1/(beta0 + nS(k,k)/2)));
       end
    elseif strcmp('Factor', TypeCov)
        O = CovFactorLoading(E);
    else 
    %Sample tau_e for iid (default)
        e1 = a0 + 0.5*n*d;
        f1 = b0 + 0.5*sum(sum(E.^2));
        tau_e = max(1e-3, gamrnd(e1,1./f1));
        O = tau_e*eye(d);
    end
%%
%%        
    % Collect samples
    if iter > nBurnin
        idx = ceil(iter-nBurnin);  
        Output.U(:,:,idx) = U;
        Output.V(:,:,idx) = V;
        Output.Delta(:,:,idx) = Delta;
%         Output.B(:,:,idx) = U*diag(Delta)*V';
        Output.G(:,:,idx) = G;

        SumDelta = SumDelta + Delta;
        SumB = SumB + B;
        SumSB = SumSB + B.^2;
        SumU = SumU + U;
        SumSU = SumSU + U.^2;
        SumV = SumV + V;
        SumSV = SumSV + V.^2;
        SumG = SumG + G;
        SumSG = SumSG + G.^2;

        SumO = SumO + O;
        SumSO = SumSO + O.^2;
    end
%     disp(iter);
 %%
end

N0 = nCollect;
Output.PostU = SumU/N0;
Output.PostV = SumV/N0;
Output.PostDelta = SumDelta/N0;
Output.PostB = SumB/N0;
Output.PostG = SumG/N0;
Output.PostO = SumO/N0;

Output.Usd = sqrt(SumSU/N0 - Output.PostU.^2);
Output.Vsd = sqrt(SumSV/N0 - Output.PostV.^2);
Output.Osd = sqrt(SumSO/N0 - Output.PostO.^2);
Output.Bsd = sqrt(SumSB/N0 - Output.PostB.^2);
Output.Gsd = sqrt(SumSG/N0 - Output.PostG.^2);


Output.nBurnin = nBurnin;
Output.nCollect = N0;

end
